//-----------------------------------------------------------------------------
// 0. INCLUDE common functions
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// 1. DRIVER SETUP - create driver object, midi ports and detection information
//-----------------------------------------------------------------------------

var midiremote_api = require('midiremote_api_v1')

var deviceDriver = midiremote_api.makeDeviceDriver('HexFM', 'NAVI WIN', 'Fabio Marchionni')

var midiInput = deviceDriver.mPorts.makeMidiInput()
var midiOutput = deviceDriver.mPorts.makeMidiOutput()

deviceDriver.makeDetectionUnit().detectPortPair(midiInput, midiOutput)
    .expectInputNameEquals('NAVI In')
    .expectOutputNameEquals('NAVI Out')


// (mc) WHY all these pairings? You have many in/out ports on your controller?
// If not, you should just keep one of them (probably the very first one above)
deviceDriver.makeDetectionUnit().detectPortPair(midiInput, midiOutput)
    .expectInputNameEquals('MIDIIN (NAVI)')
    .expectOutputNameEquals('MIDIOUT (NAVI)')

deviceDriver.makeDetectionUnit().detectPortPair(midiInput, midiOutput)
    .expectInputNameEquals('MIDIIN (NAVI) 2')
    .expectOutputNameEquals('MIDIOUT (NAVI) 2')

deviceDriver.makeDetectionUnit().detectPortPair(midiInput, midiOutput)
    .expectInputNameEquals('MIDIIN (NAVI) 3')
    .expectOutputNameEquals('MIDIOUT (NAVI) 3')


var surface = deviceDriver.mSurface





//-----------------------------------------------------------------------------
// 2. SURFACE LAYOUT - create control elements and midi bindings
//-----------------------------------------------------------------------------


// NOTE FOR SELF:
//
// MIDI Channels are base 0 (0-15 and not 1-16)
// NOTES are base 1 (1-127) no idea what Note 0 does, to be investigated


var PitchBendFader = surface.makePitchBend(0.3, 1, 0.4, 6)
var ModFader = surface.makeFader(1, 1, 1, 6)
var ExpressionFader = surface.makeFader(2, 1, 1, 6)
var VolumeFader = surface.makeFader(3, 1, 1, 6)
var PanFader = surface.makeFader(0, 0, 4, 1)

var RecToggle = surface.makeButton(0, 7, 1, 1)
var MonToggle = surface.makeButton(1, 7, 1, 1)
var ReadToggle = surface.makeButton(2, 7, 1, 1)
var WriteToggle = surface.makeButton(3, 7, 1, 1)

var LeftLocTrigger = surface.makeTriggerPad(4, 4, 1, 1)
var LocToSelTrigger = surface.makeTriggerPad(5, 4, 1, 1)
var RightLocTrigger = surface.makeTriggerPad(6, 4, 1, 1)

var SelToNewFolder = surface.makeTriggerPad(8, 4, 1, 1)
var SelToSfxChannel = surface.makeTriggerPad(9, 4, 1, 1)
var SelTGroupChannel = surface.makeTriggerPad(10, 4, 1, 1)
var RenameTrack = surface.makeTriggerPad(11, 4, 1, 1)

var ToggleVisibility = surface.makeTriggerPad(8, 5, 1, 1)
var HideUnusedMIDITracks = surface.makeTriggerPad(9, 5, 1, 1)
var ShowAllTracks = surface.makeTriggerPad(10, 5, 1, 1)
var SelToCycleMarker = surface.makeTriggerPad(11, 5, 1, 1)

// Complete with Pad (8, 6)
// Complete with Pad (9, 6)
var OpenInstrument = surface.makeTriggerPad(10, 6, 1, 1)
// Complete with Pad (11, 6)

var NewMIDITrack = surface.makeTriggerPad(10, 7, 1, 1)
var NewAudioTrack = surface.makeTriggerPad(11, 7, 1, 1)
var NewInstTrack = surface.makeTriggerPad(12, 7, 1, 1)
// Complete with Pad (13, 7)


// Arrays of elements
var SetMarker = {}
var GoToMarker = {}
var SetColor = {}

function MakeVerticalFaders (Nfaders, x1, y, w, h, step, ch, cc1, ccstep) {

    var faderstrip = {}

    for (var i=0; i<Nfaders; ++i) {
        faderstrip[i] = surface.makeFader(x1+step*i, y, w, h)

        faderstrip[i].mSurfaceValue.mMidiBinding
        .setInputPort(midiInput)
        .setOutputPort(midiOutput)
        .bindToControlChange(ch, cc1+i*ccstep)

    }

    return faderstrip

}


var QCFader = MakeVerticalFaders(8, 4, 0, 1, 4, 1, 15, 1, 1)


// ----- MARKER PADS


// Rewrite this as a function + midibinding method?
var NumOfPads = 9
var MarkerPadX = 4
var MarkerPadY = 7
var MarkerPadN = 3
var MarkerPadStep = -1

for (var i=0; i<NumOfPads; ++i) {

    GoToMarker[i] = surface.makeTriggerPad(MarkerPadX+(i % MarkerPadN), MarkerPadY + Math.floor(i/MarkerPadN)*MarkerPadStep, 1, 1)
    SetMarker[i] = surface.makeTriggerPad(MarkerPadX+(i % MarkerPadN)+0.1, MarkerPadY + Math.floor(i/MarkerPadN)*MarkerPadStep+0.1, 0.3, 0.3)

    GoToMarker[i].mSurfaceValue.mMidiBinding
        .setInputPort(midiInput)
        .bindToNote(15, i+81)

    SetMarker[i].mSurfaceValue.mMidiBinding
        .setInputPort(midiInput)
        .bindToNote(15, i+91)
}



// ------ PAGER BLANK PADS

var Blank = {}

for (var i=0; i<4; ++i) {
    Blank[i] = surface.makeBlindPanel(7, 4+i, 1, 1)
}


// ----- COLOR PADS


// Rewrite this as a function + midibinding method?
var NumOfColors = 16
var ColorPadX = 8
var ColorPadY = 4
var ColorPadN = 4
var ColorPadStep = 1

for (var i=0; i<NumOfColors; ++i) {
    SetColor[i] = surface.makeTriggerPad(ColorPadX + Math.floor(i/ColorPadN)*ColorPadStep, ColorPadY+(i % ColorPadN), 1, 1)
    
    SetColor[i].mSurfaceValue.mMidiBinding
        .setInputPort(midiInput)
        .bindToNote(15, i+11)
}


ModFader.setTypeVertical()
ExpressionFader.setTypeVertical()
VolumeFader.setTypeVertical()
PanFader.setTypeHorizontal()


PitchBendFader.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToPitchBend(1)   // MIDI Ch 2, PitchBend
ModFader.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToControlChange(0, 1)  // MIDI Ch 1, CC 1
ExpressionFader.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToControlChange(0, 11) // MIDI Ch 1, CC 11
VolumeFader.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToControlChange(0, 7)  // MIDI Ch 1, CC 7
PanFader.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToControlChange(0, 10)  // MIDI Ch 1, CC 10

RecToggle.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToNote(15, 1)  // MIDI Ch 16, Note 1
MonToggle.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToNote(15, 2)  // MIDI Ch 16, Note 2
ReadToggle.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToNote(15, 3) // MIDI Ch 16, Note 3
WriteToggle.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToNote(15, 4) // MIDI Ch 16, Note 4

LeftLocTrigger.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 5)
LocToSelTrigger.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 6)
RightLocTrigger.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 7)

SelToNewFolder.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 40)
SelToSfxChannel.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 41)
SelTGroupChannel.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 42)
RenameTrack.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 43)

ToggleVisibility.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 44)
HideUnusedMIDITracks.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 45)
ShowAllTracks.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 46)
SelToCycleMarker.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 47)

// Pad (8, 6)
// Pad (9, 6)
OpenInstrument.mSurfaceValue.mMidiBinding.setInputPort(midiInput).setOutputPort(midiOutput).bindToNote(15, 50)
// Pad (11, 6)

NewMIDITrack.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 52)
NewAudioTrack.mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 53)
NewInstTrack .mSurfaceValue.mMidiBinding.setInputPort(midiInput).bindToNote(15, 54)
// Pad (13, 7)



//-----------------------------------------------------------------------------
// 3. HOST MAPPING - create mapping pages and host bindings
//-----------------------------------------------------------------------------


// create at least one mapping page
var pageSelectedTrack = deviceDriver.mMapping.makePage('Selected Track')

// create host accessing objects
var hostSelectedTrackChannel = pageSelectedTrack.mHostAccess.mTrackSelection.mMixerChannel
var hostSelectedTrackQuickControls = pageSelectedTrack.mHostAccess.mTrackSelection.mMixerChannel.mQuickControls

// SysEx to pass to TouchOSC contains a 4-byte ID:

// COLR (0x43, 0x4F, 0x4C, 0x52) for Track Color messages
// NAME (0x4E, 0x41, 0x4D, 0x45) for Track Name messages
// TKCG (0x54, 0x4b, 0x43, 0x47) which at the moment is sent in the mOnTitleChange anyway
// QC0x (0x51, 0x43, 0x30, 0xXX) for QC Name Changes for QC1 - QC8
// QCCL (0x51, 0x43, 0x43, 0x4C) Quick Controls Clear


pageSelectedTrack.makeValueBinding(VolumeFader.mSurfaceValue, hostSelectedTrackChannel.mValue.mVolume)
pageSelectedTrack.makeValueBinding(PanFader.mSurfaceValue, hostSelectedTrackChannel.mValue.mPan)


for (var i=0; i<8; ++i) {
    //(mc) changed a bit here
    var quickControl=pageSelectedTrack.mHostAccess.mTrackSelection.mMixerChannel.mQuickControls.getByIndex(i)

    pageSelectedTrack.makeValueBinding(QCFader[i].mSurfaceValue, quickControl)
    //end changed

    // (mc) inserted
    quickControl.mOnTitleChange = (function (activeDevice, activeMapping, title,QCreference) {
        if(title!="" && QCreference!=""){
            var qcIndex=this.i
            SysExQCLabel(activeDevice,""+qcIndex, QCreference, title)
        }
    }).bind({i})
        
    
    //this one is for the value of the control, I guess you don't need this, but I leave it here, just in case
    quickControl.mOnDisplayValueChange=(function(activeDevice,activeMapping,controlValue,arg3){
        if(controlValue!=""){
            //do what you want with this value
            //I set it to a state, because I want my display when updated to also contain this value 
            activeDevice.setState("quickcontrolvalue",controlValue)
        }
    })
    //end inserted
}

pageSelectedTrack.makeValueBinding(RecToggle.mSurfaceValue, hostSelectedTrackChannel.mValue.mRecordEnable)
pageSelectedTrack.makeValueBinding(MonToggle.mSurfaceValue, hostSelectedTrackChannel.mValue.mMonitorEnable)
pageSelectedTrack.makeValueBinding(ReadToggle.mSurfaceValue, hostSelectedTrackChannel.mValue.mAutomationRead)
pageSelectedTrack.makeValueBinding(WriteToggle.mSurfaceValue, hostSelectedTrackChannel.mValue.mAutomationWrite)

pageSelectedTrack.makeCommandBinding(LocToSelTrigger.mSurfaceValue, 'Transport', 'Locators to Selection')
pageSelectedTrack.makeCommandBinding(LeftLocTrigger.mSurfaceValue, 'Transport', 'Set Left Locator')
pageSelectedTrack.makeCommandBinding(RightLocTrigger.mSurfaceValue, 'Transport', 'Set Right Locator')


pageSelectedTrack.makeCommandBinding(SelToNewFolder.mSurfaceValue, 'Project', 'Folding: Tracks To Folder')
pageSelectedTrack.makeCommandBinding(SelToSfxChannel.mSurfaceValue, 'Mixer', 'Add Track To Selected: FX Channel')
pageSelectedTrack.makeCommandBinding(SelTGroupChannel.mSurfaceValue, 'Mixer', 'Add Track To Selected: Group Channel')
pageSelectedTrack.makeCommandBinding(RenameTrack.mSurfaceValue, 'Edit', 'Rename First Selected Track')

pageSelectedTrack.makeCommandBinding(ToggleVisibility.mSurfaceValue, 'Process Project Logical Editor', 'Toggle Visibility - Selected Tracks')
pageSelectedTrack.makeCommandBinding(HideUnusedMIDITracks.mSurfaceValue, 'Process Project Logical Editor', 'Toggle Visibility - Empty MIDI Tracks')
pageSelectedTrack.makeCommandBinding(ShowAllTracks.mSurfaceValue, 'Process Project Logical Editor', 'Show All Tracks')
pageSelectedTrack.makeCommandBinding(SelToCycleMarker.mSurfaceValue, 'Macro', 'Create Cycle Marker for Selected Events')

// Pad (8, 6)
// Pad (9, 6)
pageSelectedTrack.makeCommandBinding(OpenInstrument.mSurfaceValue, 'Edit', 'Edit VST Instrument')
// Pad (11, 6)

pageSelectedTrack.makeCommandBinding(NewMIDITrack.mSurfaceValue, 'AddTrack', 'MIDI')
pageSelectedTrack.makeCommandBinding(NewAudioTrack.mSurfaceValue, 'AddTrack', 'Audio')
pageSelectedTrack.makeCommandBinding(NewInstTrack.mSurfaceValue, 'AddTrack', 'Instrument')


var ColorString = {}

for (var i=0; i<NumOfColors; ++i) {

    function ColorPadNum(num) {
        num = num.toString();
        while (num.length < 2) num = '0' + num;
        return num
    }
    ColorString[i] = String('Color ' + ColorPadNum(i+1))

     pageSelectedTrack.makeCommandBinding(SetColor[i].mSurfaceValue, 'Process Project Logical Editor', ColorString[i])  // My color-changing PLE functions are "Color 01", "Color 02", etc.

}


for (var i=0; i<NumOfPads; ++i) {

    pageSelectedTrack.makeCommandBinding(SetMarker[i].mSurfaceValue, 'Transport', String('Set Marker ' + (i+1).toString()))
    pageSelectedTrack.makeCommandBinding(GoToMarker[i].mSurfaceValue, 'Transport', String('To Marker ' + (i+1).toString()))

    // console.log(String('Set Marker '+(i+1).toString()))

}


// ------ COLOR and TEXT specific functions, sending SysEx to TouchOSC



hostSelectedTrackChannel.mOnTitleChange = function (activeDevice, activeMapping, title) {
    
    hostSelectedTrackName = title
    
    // var tkchange = [0xF0, 0x51, 0x43, 0x43, 0x4C, 0xF7 /* Message is QCCL (Quick Controls Clear)*/];
    // midiOutput.sendMidi(activeDevice, tkchange)
    // console.log("Sending TKCG")


    var msg = [0xF0, 0x4E, 0x41, 0x4D, 0x45 /* Message is NAME */];
    for (var i = 0, len = title.length; i < len; ++i) {msg.push(title.charCodeAt(i)) }
    msg.push(0xF7)
    midiOutput.sendMidi(activeDevice, msg)
    console.log("Sending NAME: " + title)

}


//(mc) I commented this one out (mc)
//hostSelectedTrackChannel.mQuickControls.getByIndex(0).mOnTitleChange = function (a, b, c, d) { console.log("QUICK CONTROL NAME CHANGE " + c +" - " + d)}

var hostSelectedTrackName = "Selected Track"

var QCName = {}


//(mc) this function should be commented out as well
//it's empty after all, but anyway I leave it as is.
hostSelectedTrackQuickControls.mOnTitleChange = function (activeDevice, activeMapping, trackname) {



    //  THE CULPRIT OF MY DESPAIR

    /*

    hostSelectedTrackQuickControls.getByIndex(0).mOnTitleChange = function (activeDevice, activeMapping, QCreference, title) { SysExQCLabel(activeDevice, '1', QCreference, title) }
    hostSelectedTrackQuickControls.getByIndex(1).mOnTitleChange = function (activeDevice, activeMapping, QCreference, title) { SysExQCLabel(activeDevice, '2', QCreference, title) }
    hostSelectedTrackQuickControls.getByIndex(2).mOnTitleChange = function (activeDevice, activeMapping, QCreference, title) { SysExQCLabel(activeDevice, '3', QCreference, title) }
    hostSelectedTrackQuickControls.getByIndex(3).mOnTitleChange = function (activeDevice, activeMapping, QCreference, title) { SysExQCLabel(activeDevice, '4', QCreference, title) }
    hostSelectedTrackQuickControls.getByIndex(4).mOnTitleChange = function (activeDevice, activeMapping, QCreference, title) { SysExQCLabel(activeDevice, '5', QCreference, title) }
    hostSelectedTrackQuickControls.getByIndex(5).mOnTitleChange = function (activeDevice, activeMapping, QCreference, title) { SysExQCLabel(activeDevice, '6', QCreference, title) }
    hostSelectedTrackQuickControls.getByIndex(6).mOnTitleChange = function (activeDevice, activeMapping, QCreference, title) { SysExQCLabel(activeDevice, '7', QCreference, title) }
    hostSelectedTrackQuickControls.getByIndex(7).mOnTitleChange = function (activeDevice, activeMapping, QCreference, title) { SysExQCLabel(activeDevice, '8', QCreference, title) }

    */

}


/**
 * @param {MR_ActiveDevice} activeDevice
 * @param {string} QCnum
 * @param {string} QCreference
 * @param {string} title
 */

function SysExQCLabel (activeDevice, QCnum, QCreference, title) {

    
    if (QCreference != '')
        {
            // if (QCreference == hostSelectedTrackInstrumentName || QCreference == hostSelectedTrackName) {title = QCreference + " - " + title}
           

            var msg = [0xF0, 0x51, 0x43, 0x30, QCnum.charCodeAt(0)];
            // console.log(QCnum.charCodeAt(0).toString())
            for (var i = 0, len = title.length; i < len; ++i) {msg.push(title.charCodeAt(i)) }
            msg.push(0xF7)
            
            midiOutput.sendMidi(activeDevice, msg)


            console.log("Sending QC0" + QCnum + ": "  /* + msg.toString() + " | " */  + QCreference + " - " + title)

    }

}


hostSelectedTrackChannel.mOnColorChange = function(activeDevice, activeMapping, r, g, b, a, isActive) {

    var msg = [0xF0, 0x43, 0x4F, 0x4C, 0x52 /* Message is "COLR" */];

    r = Math.ceil(r*255)
    g = Math.ceil(g*255)
    b = Math.ceil(b*255)

 
    msg.push(r)
    msg.push(g)
    msg.push(b)
    msg.push(0xF7)
    
    midiOutput.sendMidi(activeDevice, msg)

    console.log("Sending COLR: (" + r.toString() + " " + g.toString() + " " + b.toString() +")")

}


